import sys
import time
from attr import dataclass
from typing import Dict, TypedDict, List
from openai import OpenAI

from gptcli.types import Message
from gptcli.openai_types import ThreadMessage, ThreadRun

class AssistantConfig(TypedDict, total=False):
    id: str
    messages: List[Message]

CONFIG_DEFAULTS = {
    "id": "asst_jCP75X9phRfVjZ8Q4iBistYT",
}

DEFAULT_ASSISTANTS: Dict[str, AssistantConfig] = {}


class AssistantThread():
    """
    A class to represent an assistant thread.

    Upon instantiation, a new thread is created against the given ChatGPT Assistant

    In future we can decouple Assistants from Threads: 
    - create an Assistant class that can contain multiple AssistantThreads.
    """
    def __init__(self, config: AssistantConfig):
        self.config = config
        self.openai_client = OpenAI()
        self.assistant_handle = self.openai_client.beta.assistants.retrieve(config.get("id"))
        self.last_user_message_id = None
        self.init_messages()

    @classmethod
    def from_config(cls, name: str, config: AssistantConfig):
        config = config.copy()
        if name in DEFAULT_ASSISTANTS:
            # Merge the config with the default config
            # If a key is in both, use the value from the config
            default_config = DEFAULT_ASSISTANTS[name]
            for key in [*config.keys(), *default_config.keys()]:
                if config.get(key) is None:
                    config[key] = default_config[key]

        return cls(config)

    def init_messages(self) -> List[Message]:
        """
        Create a new OpenAI thread and return the default messages.
        """
        self.thread = self.openai_client.beta.threads.create()  

        return self.config.get("messages", [])[:]


    def add_message(self, our_message: Message) -> ThreadMessage:
        """
        Send a message to the chatgpt Thread associated with this assistant and return the response.
        """
        their_message = self.openai_client.beta.threads.messages.create(
            thread_id=self.thread.id,
            role=our_message['role'],
            content=our_message['content'],
        )
        self.last_user_message_id = their_message.id
        return their_message

    def run_thread(self) -> ThreadRun:
        """
        Start a Run on the chatgpt Thread associated with this assistant and wait for it to complete.
        """
        run = self.openai_client.beta.threads.runs.create(
            thread_id=self.thread.id,
            assistant_id=self.assistant_handle.id,
        )
        # TODO move out of this function. Use async primitive instead.
        while run.status != "completed":
            time.sleep(2)
            run = self.openai_client.beta.threads.runs.retrieve(run.id, thread_id=self.thread.id)

        return run

    def fetch_messages(self, since_last_user_message: bool) -> List[ThreadMessage]:
        # TODO keep SyncCursorPage instead of immediately converting to list? 
        # May become a problem when threads become long enough to split into multiple pages
        messages = list(self.openai_client.beta.threads.messages.list(
            thread_id=self.thread.id
        ))
        # Messages come back in reverse chronological order. We reverse them.
        messages.reverse()

        # return all new messages (i.e. all the ones after the one we just added)
        if since_last_user_message and self.last_user_message_id:
            last_message_id = self.last_user_message_id
            message_ids = [message.id for message in messages]
            last_message_index = message_ids.index(last_message_id)
            if last_message_index == -1:
                raise ValueError("last_message_id not found in messages")
            messages = messages[last_message_index+1:]

        return self.add_citations_to_messages(messages)
    
    def add_citations_to_messages(self, messages):  
        messages_with_citations = []
        for message in messages:
            # Extract the message content
            # Assumes one text message per message
            if len(message.content) > 1:
                raise ValueError("Unimplemented: More than one text message per message")
            message_content = message.content[0].text
            annotations = message_content.annotations
            citations = []

            # Iterate over the annotations and add footnotes
            for index, annotation in enumerate(annotations):
                # Replace the text with a footnote
                message_content.value = message_content.value.replace(annotation.text, f' [{index}]')

                # Gather citations based on annotation attributes
                if (file_citation := getattr(annotation, 'file_citation', None)):
                    cited_file = self.openai_client.files.retrieve(file_citation.file_id)
                    searchable_quote = ' '.join(file_citation.quote.split()[:6])
                    citations.append(f'[{index}] {cited_file.filename} - (Search: "{searchable_quote}")')
                elif (file_path := getattr(annotation, 'file_path', None)):
                    cited_file = self.openai_client.files.retrieve(file_path.file_id)
                    citations.append(f'[{index}] Click <here> to download {cited_file.filename}')

            # Add footnotes to the end of the message before displaying to user
            message_content.value += '\n\n' + '\n'.join(citations)
            messages_with_citations.append(message_content.value)
        
        return messages

    def get_thread_id(self):
        return self.thread.id
    
    def get_assistant_id(self):
        return self.assistant_handle.id

@dataclass
class AssistantGlobalArgs:
    assistant_name: str

def init_assistant(
    args: AssistantGlobalArgs, custom_assistants: Dict[str, AssistantConfig]
) -> AssistantThread:
    name = args.assistant_name
    if name in custom_assistants:
        assistant = AssistantThread.from_config(name, custom_assistants[name])
    elif name in DEFAULT_ASSISTANTS:
        assistant = AssistantThread.from_config(name, DEFAULT_ASSISTANTS[name])
    else:
        print(f"Unknown assistant: {name}")
        sys.exit(1)

    return assistant

def thread_message_to_text(thread_messages: ThreadMessage) -> str:
    thread_messages = [content for thread_message in thread_messages for content in thread_message.content]
    thread_contents = filter(lambda content: content.type == "text", thread_messages)
    thread_texts = [content.text.value + "\n" for content in thread_contents]
    return thread_texts


"""
Since the openai python library doesn't have type annotations, we have to define our own types for the objects we get back from the API.
"""
class MessageText:
    def __init__(self, value, annotations):
        self.value = value
        self.annotations = annotations

class Content:
    def __init__(self, type, text):
        self.type = type
        self.text = text

# https://platform.openai.com/docs/api-reference/messages/object
class ThreadMessage:
    def __init__(self, id, object, created_at, thread_id, role, content, file_ids, assistant_id, run_id, metadata):
        self.id = id
        self.object = object
        self.created_at = created_at
        self.thread_id = thread_id
        self.role = role
        self.content = [Content(item['type'], MessageText(item['text']['value'], item['text']['annotations'])) for item in content]
        self.file_ids = file_ids
        self.assistant_id = assistant_id
        self.run_id = run_id
        self.metadata = metadata     

class Tool:
    def __init__(self, type):
        self.type = type

# https://platform.openai.com/docs/api-reference/runs/object
class ThreadRun:
    def __init__(self, id, object, created_at, assistant_id, thread_id, status, started_at, expires_at, cancelled_at, failed_at, completed_at, last_error, model, instructions, tools, file_ids, metadata):
        self.id = id
        self.object = object
        self.created_at = created_at
        self.assistant_id = assistant_id
        self.thread_id = thread_id
        self.status = status
        self.started_at = started_at
        self.expires_at = expires_at
        self.cancelled_at = cancelled_at
        self.failed_at = failed_at
        self.completed_at = completed_at
        self.last_error = last_error
        self.model = model
        self.instructions = instructions
        self.tools = [Tool(tool['type']) for tool in tools]
        self.file_ids = file_ids
        self.metadata = metadata



"""
This module is responsible for calculating the cost of a chat session.
"""

import tiktoken
import logging

from gptcli.types import Message
from gptcli.session import ChatListener
from gptcli.assistant import AssistantThread

from rich.console import Console
from typing import List, Optional


def num_tokens_from_messages(messages: List[Message], model: str) -> Optional[int]:
    return num_tokens_from_messages_openai(messages, model)


def num_tokens_from_completion(message: Message, model: str) -> Optional[int]:
    return num_tokens_from_completion_openai(message, model)


GPT_3_5_TURBO_PRICE_PER_TOKEN = {
    "prompt": 0.0015 / 1000,
    "response": 0.002 / 1000,
}

GPT_3_5_TURBO_16K_PRICE_PER_TOKEN = {
    "prompt": 0.003 / 1000,
    "response": 0.004 / 1000,
}

GPT_4_PRICE_PER_TOKEN = {
    "prompt": 0.03 / 1000,
    "response": 0.06 / 1000,
}

GPT_4_TURBO_PRICE_PER_TOKEN = {
    "prompt": 0.01 / 1000,
    "response": 0.03 / 1000,
}

GPT_4_32K_PRICE_PER_TOKEN = {
    "prompt": 0.06 / 1000,
    "response": 0.12 / 1000,
}


def gpt_pricing(model: str, prompt: bool) -> Optional[float]:
    if model.startswith("gpt-3.5-turbo-16k"):
        pricing = GPT_3_5_TURBO_16K_PRICE_PER_TOKEN
    elif model.startswith("gpt-3.5-turbo"):
        pricing = GPT_3_5_TURBO_PRICE_PER_TOKEN
    elif model.startswith("gpt-4-32k"):
        pricing = GPT_4_32K_PRICE_PER_TOKEN
    elif model.startswith("gpt-4-1106-preview"):
        pricing = GPT_4_TURBO_PRICE_PER_TOKEN
    elif model.startswith("gpt-4"):
        pricing = GPT_4_PRICE_PER_TOKEN
    else:
        return None
    return pricing["prompt" if prompt else "response"]


CLAUDE_PRICE_PER_TOKEN = {
    "prompt": 11.02 / 1_000_000,
    "response": 32.68 / 1_000_000,
}

CLAUDE_INSTANT_PRICE_PER_TOKEN = {
    "prompt": 1.63 / 1_000_000,
    "response": 5.51 / 1_000_000,
}


def claude_pricing(model: str, prompt: bool) -> Optional[float]:
    if "instant" in model:
        pricing = CLAUDE_INSTANT_PRICE_PER_TOKEN
    else:
        pricing = CLAUDE_PRICE_PER_TOKEN
    return pricing.get("prompt" if prompt else "response")


def price_per_token(model: str, prompt: bool) -> Optional[float]:
    if model.startswith("gpt"):
        return gpt_pricing(model, prompt)
    elif model.startswith("claude"):
        return claude_pricing(model, prompt)
    elif model.startswith("chat-bison"):
        return 0
    else:
        return None


def price_for_completion(messages: List[Message], response: Message, model: str):
    num_tokens_prompt = num_tokens_from_messages(messages, model)
    num_tokens_response = num_tokens_from_completion(response, model)
    if num_tokens_prompt is None or num_tokens_response is None:
        return None

    price_per_token_prompt = price_per_token(model, prompt=True)
    price_per_token_response = price_per_token(model, prompt=False)
    if price_per_token_prompt is None or price_per_token_response is None:
        return None

    return (
        price_per_token_prompt * num_tokens_prompt
        + price_per_token_response * num_tokens_response
    )

class PriceChatListener(ChatListener):
    def __init__(self, assistant: AssistantThread):
        self.assistant = assistant
        self.current_spend = 0
        self.logger = logging.getLogger("gptcli-price")
        self.console = Console()

    def on_chat_clear(self):
        self.current_spend = 0

    def on_chat_response(
        self, messages: List[Message], response: Message
    ):
        model = self.assistant._param("model")
        num_tokens = num_tokens_from_messages(messages + [response], model)
        price = price_for_completion(messages, response, model)
        if price is None:
            self.logger.error(f"Cannot get cost information for model {model}")
            return
        self.current_spend += price
        self.logger.info(f"Token usage {num_tokens}")
        self.logger.info(f"Message price (model: {model}): ${price:.3f}")
        self.logger.info(f"Current spend: ${self.current_spend:.3f}")
        self.console.print(
            f"Tokens: {num_tokens} | Price: ${price:.3f} | Total: ${self.current_spend:.3f}",
            justify="right",
            style="dim",
        )

def num_tokens_from_messages_openai(messages: List[Message], model: str) -> int:
    encoding = tiktoken.encoding_for_model(model)
    num_tokens = 0
    for message in messages:
        # every message follows <im_start>{role/name}\n{content}<im_end>\n
        num_tokens += 4
        for key, value in message.items():
            assert isinstance(value, str)
            num_tokens += len(encoding.encode(value))
            if key == "name":  # if there's a name, the role is omitted
                num_tokens += -1  # role is always required and always 1 token
    num_tokens += 2  # every reply is primed with <im_start>assistant
    return num_tokens


def num_tokens_from_completion_openai(completion: Message, model: str) -> int:
    return num_tokens_from_messages_openai([completion], model)


from unittest import mock
import httpx

from openai import BadRequestError, OpenAIError

from gptcli.session import ChatSession
from gptcli.openai_types import ThreadMessage


def setup_assistant_mock():
    assistant_mock = mock.MagicMock()
    assistant_mock.init_messages.return_value = []
    return assistant_mock


def setup_listener_mock():
    listener_mock = mock.MagicMock()
    response_streamer_mock = mock.MagicMock()
    response_streamer_mock.__enter__.return_value = response_streamer_mock
    listener_mock.response_streamer.return_value = response_streamer_mock
    return listener_mock, response_streamer_mock


def setup_session():
    assistant_mock = setup_assistant_mock()
    listener_mock, _ = setup_listener_mock()
    session = ChatSession(assistant_mock, listener_mock)
    return assistant_mock, listener_mock, session

def create_thread_message(role, content):
    return ThreadMessage(
        id="msg_feBzkJcQlHRyg4BsLm9GHwwf",
        assistant_id="asst_jCP75X9phRfVjZ8Q4iBistYT",
        content=[
            {
                "text": {
                    "annotations": [],
                    "value": content
                },
                "type": "text"
            }
        ],
        created_at=1703532865,
        file_ids=[],
        metadata={},
        object="thread.message",
        role=role,
        run_id="run_ZuVSExmplLDv2Z76FJNRXEzw",
        thread_id="thread_RSkXNj7NpXmAdgUTDc8Fm3XX"
    )


def test_simple_input():
    assistant_mock, listener_mock, session = setup_session()

    expected_response = "assistant message"
    assistant_mock.fetch_messages.return_value = [
        create_thread_message("assistant", expected_response)
    ]

    user_input = "user message"
    should_continue = session.process_input(user_input, {})
    assert should_continue

    user_message = {"role": "user", "content": user_input}
    assistant_message = {"role": "assistant", "content": expected_response}

    assistant_mock.add_message.assert_called_once_with(user_message)
    assistant_mock.run_thread.assert_called_once()
    assistant_mock.fetch_messages.assert_called_once()
    listener_mock.on_chat_message.assert_has_calls(
        [mock.call(user_message), mock.call(assistant_message)]
    )


def test_quit():
    _, _, session = setup_session()
    should_continue = session.process_input(":q", {})
    assert not should_continue

def test_clear():
    user_input = "user message"
    assistant_mock, listener_mock, session = setup_session()

    assistant_mock.init_messages.assert_called_once()
    assistant_mock.init_messages.reset_mock()

    assistant_mock.fetch_messages.return_value = [
        create_thread_message("assistant", "assistant_message")
    ]   

    should_continue = session.process_input(user_input, {})
    assert should_continue

    assistant_mock.add_message.assert_called_once_with(
        {"role": "user", "content": user_input}
    )
    listener_mock.on_chat_message.assert_has_calls(
        [
            mock.call({"role": "user", "content": user_input}),
            mock.call({"role": "assistant", "content": "assistant_message"}),
        ]
    )
    assistant_mock.add_message.reset_mock()
    assistant_mock.fetch_messages.reset_mock()
    listener_mock.on_chat_message.reset_mock()

    should_continue = session.process_input(":c", {})
    assert should_continue

    assistant_mock.init_messages.assert_called_once()
    listener_mock.on_chat_clear.assert_called_once()
    assistant_mock.add_message.assert_not_called()

    assistant_output = "assistant message 2"
    user_input_2 = "user message 2"
    assistant_mock.fetch_messages.return_value = [
        create_thread_message("assistant", assistant_output)
    ]

    should_continue = session.process_input(user_input_2, {})
    assert should_continue

    assistant_mock.add_message.assert_called_once_with(
        {"role": "user", "content": user_input_2}
    )
    listener_mock.on_chat_message.assert_has_calls(
        [
            mock.call({"role": "user", "content": user_input_2}),
            mock.call({"role": "assistant", "content": assistant_output}),
        ]
    )


def test_rerun():
    assistant_mock, listener_mock, session = setup_session()

    assistant_mock.init_messages.assert_called_once()
    assistant_mock.init_messages.reset_mock()

    # Re-run before any input shouldn't do anything
    should_continue = session.process_input(":r", {})
    assert should_continue

    assistant_mock.init_messages.assert_not_called()
    assistant_mock.complete_chat.assert_not_called()
    listener_mock.on_chat_message.assert_not_called()
    listener_mock.on_chat_rerun.assert_called_once_with(False)

    listener_mock.on_chat_rerun.reset_mock()

    # Now proper re-run
    assistant_output = "assistant message"
    user_input = "user message"
    assistant_mock.fetch_messages.return_value = [
        create_thread_message("assistant", assistant_output)
    ]

    should_continue = session.process_input(user_input, {})
    assert should_continue

    assistant_mock.add_message.assert_called_once_with(
        {"role": "user", "content": user_input}
    )
    listener_mock.on_chat_message.assert_has_calls(
        [
            mock.call({"role": "user", "content": user_input}),
            mock.call({"role": "assistant", "content": assistant_output}),
        ]
    )
    assistant_mock.run_thread.reset_mock()
    assistant_mock.add_message.reset_mock()
    assistant_mock.fetch_messages.reset_mock()
    listener_mock.on_chat_message.reset_mock()

    assistant_output_after_reset = "assistant message after reste"
    assistant_mock.fetch_messages.return_value = [
        create_thread_message("assistant", assistant_output_after_reset)
    ]

    should_continue = session.process_input(":r", {})
    assert should_continue

    listener_mock.on_chat_rerun.assert_called_once_with(True)
    assistant_mock.run_thread.assert_called_once()
    assistant_mock.fetch_messages.assert_called_once()

    listener_mock.on_chat_message.assert_has_calls(
        [
            mock.call({"role": "assistant", "content": assistant_output_after_reset}),
        ]
    )


def test_invalid_request_error():
    assistant_mock, listener_mock, session = setup_session()

    error = BadRequestError(
        "error message",
        response=httpx.Response(
            401, request=httpx.Request("POST", "http://localhost/")
        ),
        body=None,
    )
    assistant_mock.run_thread.side_effect = error

    user_input = "user message"
    should_continue = session.process_input(user_input, {})
    assert should_continue

    user_message = {"role": "user", "content": user_input}
    listener_mock.on_chat_message.assert_has_calls([mock.call(user_message)])
    listener_mock.on_error.assert_called_once_with(error)
    print("done")
    # Now rerun shouldn't do anything because user input was not saved
    assistant_mock.run_thread.reset_mock()
    listener_mock.on_chat_message.reset_mock()
    listener_mock.on_error.reset_mock()

    should_continue = session.process_input(":r", {})
    assert should_continue

    assistant_mock.run_thread.assert_not_called()
    listener_mock.on_chat_message.assert_not_called()
    listener_mock.on_error.assert_not_called()
    listener_mock.on_chat_rerun.assert_called_once_with(False)


class OpenAITestError(OpenAIError):
    pass


def test_openai_error():
    assistant_mock, listener_mock, session = setup_session()

    error = OpenAITestError()
    assistant_mock.run_thread.side_effect = error

    user_input = "user message"
    should_continue = session.process_input(user_input, {})
    assert should_continue

    user_message = {"role": "user", "content": user_input}
    listener_mock.on_chat_message.assert_has_calls([mock.call(user_message)])
    listener_mock.on_error.assert_called_once_with(error)

    # Re-run should work
    assistant_mock.run_thread.reset_mock()
    listener_mock.on_chat_message.reset_mock()
    listener_mock.on_error.reset_mock()

    assistant_mock.run_thread.side_effect = None
    assistant_output = "assistant message"
    assistant_message = {"role": "assistant", "content": assistant_output}
    assistant_mock.fetch_messages.return_value = [
        create_thread_message("assistant", assistant_output)
    ]

    should_continue = session.process_input(":r", {})
    assert should_continue

    assistant_mock.add_message.assert_called_once_with(user_message)
    listener_mock.on_chat_message.assert_has_calls(
        [
            mock.call(assistant_message),
        ]
    )



from gptcli.session import ChatListener
from gptcli.types import Message
from pathlib import Path

# TODO use the logging package to do writes instead of our own class

class PersistChatListener(ChatListener):
    def __init__(self, thread_id, assistant_id: str):
        # create a file with the thread id as the name in the current directory
        Path("./logs").mkdir(exist_ok=True)
        self.file_handle = open(f"./logs/gptcli-{assistant_id}-{thread_id}.log", "w")

    def on_chat_start(self):
        self.file_handle.write("Chat started.\n")


    def on_chat_clear(self):
        self.file_handle.write("Cleared the conversation.\n")
        
    def on_chat_rerun(self, success: bool):
        if success:
            self.file_handle.write("Re-generating the last message.\n")

    def on_error(self, e: Exception):
        self.file_handle.write(str(e))
        self.file_handle.close()

    def on_chat_message(self, message: Message):
        self.file_handle.write(f"{message['role']}: {message['content']}\n")

    def on_chat_end(self):
        self.file_handle.close()


import logging
from gptcli.types import Message
from gptcli.session import ChatListener


class LoggingChatListener(ChatListener):
    def __init__(self):
        self.logger = logging.getLogger("gptcli-session")

    def on_chat_start(self):
        self.logger.info("Chat started")

    def on_chat_clear(self):
        self.logger.info("Cleared the conversation.")

    def on_chat_rerun(self, success: bool):
        if success:
            self.logger.info("Re-generating the last message.")

    def on_error(self, e: Exception):
        self.logger.exception(e)

    def on_chat_message(self, message: Message):
        self.logger.info(f"{message['role']}: {message['content']}")



# from gptcli.cli import parse_args


# def test_parse_args():
#     assert parse_args("foo") == ("foo", {})
#     assert parse_args("foo bar") == ("foo bar", {})
#     assert parse_args("this is a prompt --bar 1.0") == (
#         "this is a prompt",
#         {"bar": "1.0"},
#     )
#     assert parse_args("this is a prompt --bar 1.0 --baz    2.0") == (
#         "this is a prompt",
#         {"bar": "1.0", "baz": "2.0"},
#     )
#     assert parse_args("this is a prompt --bar=1.0 --baz=2.0") == (
#         "this is a prompt",
#         {"bar": "1.0", "baz": "2.0"},
#     )




from gptcli.types import Message
from gptcli.session import ChatListener, ResponseStreamer


from typing import List


class CompositeResponseStreamer(ResponseStreamer):
    def __init__(self, streamers: List[ResponseStreamer]):
        self.streamers = streamers

    def __enter__(self):
        for streamer in self.streamers:
            streamer.__enter__()
        return self

    def on_next_token(self, token: str):
        for streamer in self.streamers:
            streamer.on_next_token(token)

    def __exit__(self, *args):
        for streamer in self.streamers:
            streamer.__exit__(*args)


class CompositeChatListener(ChatListener):
    def __init__(self, listeners: List[ChatListener]):
        self.listeners = listeners

    def on_chat_start(self):
        for listener in self.listeners:
            listener.on_chat_start()

    def on_chat_clear(self):
        for listener in self.listeners:
            listener.on_chat_clear()

    def on_chat_rerun(self, success: bool):
        for listener in self.listeners:
            listener.on_chat_rerun(success)

    def on_error(self, e: Exception):
        for listener in self.listeners:
            listener.on_error(e)

    def response_streamer(self) -> ResponseStreamer:
        return CompositeResponseStreamer(
            [listener.response_streamer() for listener in self.listeners]
        )

    def on_chat_message(self, message: Message):
        for listener in self.listeners:
            listener.on_chat_message(message)

    def on_chat_response(
        self, messages: List[Message], response: Message
    ):
        for listener in self.listeners:
            listener.on_chat_response(messages, response)
    
    def on_chat_end(
        self
    ):
        for listener in self.listeners:
            listener.on_chat_end()



import re
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.key_binding.bindings import named_commands
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from typing import Any, Dict, Optional, Tuple

from rich.text import Text
from gptcli.session import (
    ALL_COMMANDS,
    COMMAND_CLEAR,
    COMMAND_QUIT,
    COMMAND_RERUN,
    ChatListener,
    InvalidArgumentError,
    ResponseStreamer,
    UserInputProvider,
)


TERMINAL_WELCOME = """
Hi! I'm here to help. Type `:q` or Ctrl-D to exit, `:c` or Ctrl-C and Enter to clear
the conversation, `:r` or Ctrl-R to re-generate the last response.
To enter multi-line mode, enter a backslash `\\` followed by a new line.
Exit the multi-line mode by pressing ESC and then Enter (Meta+Enter).
Try `:?` for help.
"""


class StreamingMarkdownPrinter:
    def __init__(self, console: Console, markdown: bool):
        self.console = console
        self.current_text = ""
        self.markdown = markdown
        self.live: Optional[Live] = None

    def __enter__(self) -> "StreamingMarkdownPrinter":
        if self.markdown:
            self.live = Live(
                console=self.console, auto_refresh=False, vertical_overflow="visible"
            )
            self.live.__enter__()
        return self

    def print(self, text: str):
        self.current_text += text
        if self.markdown:
            assert self.live
            content = Markdown(self.current_text, style="green")
            self.live.update(content)
            self.live.refresh()
        else:
            self.console.print(Text(text, style="green"), end="")

    def __exit__(self, *args):
        if self.markdown:
            assert self.live
            self.live.__exit__(*args)
        self.console.print()


class CLIResponseStreamer(ResponseStreamer):
    def __init__(self, console: Console, markdown: bool):
        self.console = console
        self.markdown = markdown
        self.printer = StreamingMarkdownPrinter(self.console, self.markdown)
        self.first_token = True

    def __enter__(self):
        self.printer.__enter__()
        return self

    def on_next_token(self, token: str):
        if self.first_token and token.startswith(" "):
            token = token[1:]
        self.first_token = False
        self.printer.print(token)

    def __exit__(self, *args):
        self.printer.__exit__(*args)


class CLIChatListener(ChatListener):
    def __init__(self, markdown: bool):
        self.markdown = markdown
        self.console = Console()

    def on_chat_start(self):
        console = Console(width=80)
        console.print(Markdown(TERMINAL_WELCOME))

    def on_chat_clear(self):
        self.console.print("[bold]Cleared the conversation.[/bold]")

    def on_chat_rerun(self, success: bool):
        if success:
            self.console.print("[bold]Re-running the last message.[/bold]")
        else:
            self.console.print("[bold]Nothing to re-run.[/bold]")

    def on_error(self, e: Exception):
        if isinstance(e, InvalidArgumentError):
            self.console.print(f"[red]{e.message}[/red]")
        else:
            self.console.print(f"[red]Error: {type(e)}: {e}[/red]")

    def response_streamer(self) -> ResponseStreamer:
        return CLIResponseStreamer(self.console, self.markdown)


def parse_args(input: str) -> Tuple[str, Dict[str, Any]]:
    args = {}
    regex = r"--(\w+)(?:\s+|=)([^\s]+)"
    matches = re.findall(regex, input)
    if matches:
        args = dict(matches)
        input = input.split("--")[0].strip()

    return input, args


class CLIFileHistory(FileHistory):
    def append_string(self, string: str) -> None:
        if string in ALL_COMMANDS:
            return
        return super().append_string(string)


class CLIUserInputProvider(UserInputProvider):
    def __init__(self, history_filename) -> None:
        self.prompt_session = PromptSession[str](
            history=CLIFileHistory(history_filename)
        )

    def get_user_input(self) -> Tuple[str, Dict[str, Any]]:
        while (next_user_input := self._request_input()) == "":
            pass

        user_input, args = self._parse_input(next_user_input)
        return user_input, args

    def prompt(self, multiline=False):
        bindings = KeyBindings()

        bindings.add("c-a")(named_commands.get_by_name("beginning-of-line"))
        bindings.add("c-b")(named_commands.get_by_name("backward-char"))
        bindings.add("c-e")(named_commands.get_by_name("end-of-line"))
        bindings.add("c-f")(named_commands.get_by_name("forward-char"))
        bindings.add("c-left")(named_commands.get_by_name("backward-word"))
        bindings.add("c-right")(named_commands.get_by_name("forward-word"))

        @bindings.add("c-c")
        def _(event: KeyPressEvent):
            if len(event.current_buffer.text) == 0 and not multiline:
                event.current_buffer.text = COMMAND_CLEAR[0]
                event.current_buffer.cursor_right(len(COMMAND_CLEAR[0]))
            else:
                event.app.exit(exception=KeyboardInterrupt, style="class:aborting")

        @bindings.add("c-d")
        def _(event: KeyPressEvent):
            if len(event.current_buffer.text) == 0:
                if not multiline:
                    event.current_buffer.text = COMMAND_QUIT[0]
                event.current_buffer.validate_and_handle()

        @bindings.add("c-r")
        def _(event: KeyPressEvent):
            if len(event.current_buffer.text) == 0:
                event.current_buffer.text = COMMAND_RERUN[0]
                event.current_buffer.validate_and_handle()

        try:
            return self.prompt_session.prompt(
                "> " if not multiline else "multiline> ",
                vi_mode=True,
                multiline=multiline,
                enable_open_in_editor=True,
                key_bindings=bindings,
            )
        except KeyboardInterrupt:
            return ""

    def _request_input(self):
        line = self.prompt()

        if line != "\\":
            return line

        return self.prompt(multiline=True)

    def _parse_input(self, input: str) -> Tuple[str, Dict[str, Any]]:
        input, args = parse_args(input)
        return input, args



from abc import abstractmethod
from openai import BadRequestError, OpenAIError
from gptcli.types import Message
from typing import Any, Dict, List, Tuple
from gptcli.assistant import AssistantThread, thread_message_to_text

class ResponseStreamer:
    def __enter__(self) -> "ResponseStreamer":
        return self

    def on_next_token(self, token: str):
        pass

    def __exit__(self, *args):
        pass


class ChatListener:
    def on_chat_start(self):
        pass

    def on_chat_clear(self):
        pass

    def on_chat_rerun(self, success: bool):
        pass

    def on_error(self, error: Exception):
        pass

    def response_streamer(self) -> ResponseStreamer:
        return ResponseStreamer()

    def on_chat_message(self, message: Message):
        pass

    def on_chat_response(
        self, messages: List[Message], response: Message
    ):
        pass

    def on_chat_end(
        self
    ):
        pass


class UserInputProvider:
    @abstractmethod
    def get_user_input(self) -> Tuple[str, Dict[str, Any]]:
        pass


class InvalidArgumentError(Exception):
    def __init__(self, message: str):
        self.message = message


COMMAND_CLEAR = (":clear", ":c")
COMMAND_QUIT = (":quit", ":q")
COMMAND_RERUN = (":rerun", ":r")
COMMAND_HELP = (":help", ":h", ":?")
ALL_COMMANDS = [*COMMAND_CLEAR, *COMMAND_QUIT, *COMMAND_RERUN, *COMMAND_HELP]
COMMANDS_HELP = """
Commands:
- `:clear` / `:c` / Ctrl+C - Clear the conversation.
- `:quit` / `:q` / Ctrl+D - Quit the program.
- `:rerun` / `:r` / Ctrl+R - Re-run the last message.
- `:help` / `:h` / `:?` - Show this help message.
"""

class ChatSession:
    # This class represents a single CLI session. Including the assistant and messages between it and the user.
    def __init__(
        self,
        assistant: AssistantThread,
        listener: ChatListener,
    ):
        self.assistant = assistant
        self.messages: List[Message] = assistant.init_messages()
        self.user_prompts: List[Message] = []
        self.listener = listener

    def _clear(self):
        self.messages = self.assistant.init_messages()
        self.user_prompts = []
        self.listener.on_chat_clear()

    def _rerun(self):
        if len(self.user_prompts) == 0:
            self.listener.on_chat_rerun(False)
            return

        if self.messages[-1]["role"] == "assistant":
            self.messages = self.messages[:-1]

        self.listener.on_chat_rerun(True)
        self._get_response()

    def _get_response(self) -> bool:
        """
        Respond to the user's input and return whether the assistant's response was saved.
        """
        next_response: str = ""
        try:
            self.assistant.run_thread()
            # Fetch the text of all recent messages
            thread_messages = self.assistant.fetch_messages(since_last_user_message=True)
            thread_texts = thread_message_to_text(thread_messages)

            with self.listener.response_streamer() as stream:
                for response in thread_texts:
                    next_response += response
                    stream.on_next_token(response)
        except KeyboardInterrupt:
            # If the user interrupts the response, we'll just return what we have so far
            pass
        except BadRequestError as e:
            self.listener.on_error(e)
            return False
        except OpenAIError as e:
            self.listener.on_error(e)
            return True

        response_message: Message = {"role": "assistant", "content": next_response}
        self.listener.on_chat_message(response_message)
        self.listener.on_chat_response(self.messages, response_message)

        self.messages = self.messages + [response_message]
        return True

    def _add_user_message(self, user_input: str) -> Message:
        user_message: Message = {"role": "user", "content": user_input}
        self.assistant.add_message(user_message)
        self.messages = self.messages + [user_message]
        self.listener.on_chat_message(user_message)
        self.user_prompts.append(user_message)
        return user_message

    def _rollback_user_message(self):
        self.messages = self.messages[:-1]
        self.user_prompts = self.user_prompts[:-1]

    def _print_help(self):
        with self.listener.response_streamer() as stream:
            stream.on_next_token(COMMANDS_HELP)

    def _quit(self):
        self.listener.on_chat_end()

    def process_input(self, user_input: str, args: Dict[str, Any]):
        """
        Process the user's input and return whether the session should continue.
        """
        if not self._validate_args(args):
            return True

        if user_input in COMMAND_QUIT:
            self._quit()
            return False
        elif user_input in COMMAND_CLEAR:
            self._clear()
            return True
        elif user_input in COMMAND_RERUN:
            self._rerun()
            return True
        elif user_input in COMMAND_HELP:
            self._print_help()
            return True

        self._add_user_message(user_input)
        response_saved = self._get_response()
        if not response_saved:
            self._rollback_user_message()

        return True

    def _validate_args(self, args: Dict[str, Any]) -> bool:
        for key in args:
            supported_overrides = self.assistant.supported_overrides()
            if key not in supported_overrides:
                self.listener.on_error(
                    InvalidArgumentError(
                        f"Invalid argument: {key}. Allowed arguments: {supported_overrides}"
                    )
                )
                return False
        return True
    
    def loop(self, input_provider: UserInputProvider):
        self.listener.on_chat_start()
        while self.process_input(*input_provider.get_user_input()):
            pass



#!/usr/bin/env python

"""
This is the main entry point for the gpt-cli program.
"""

import sys

import gptcli

MIN_PYTHON = (3, 9)
if sys.version_info < MIN_PYTHON:
    sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

import os
from typing import cast
import argparse
import sys


from gptcli.assistant import (
    AssistantThread,
    DEFAULT_ASSISTANTS,
    AssistantGlobalArgs,
    init_assistant,
)
from gptcli.cli import (
    CLIChatListener,
    CLIUserInputProvider,
)
from gptcli.composite import CompositeChatListener
from gptcli.config import (
    CONFIG_FILE_PATHS,
    GptCliConfig,
    choose_config_file,
    read_yaml_config,
)
from gptcli.logging_utils import LoggingChatListener
from gptcli.persist import PersistChatListener
from gptcli.cost import PriceChatListener
from gptcli.session import ChatSession

default_exception_handler = sys.excepthook


def exception_handler(type, value, traceback):
    print("An uncaught exception occurred. Please report this issue on GitHub.")
    default_exception_handler(type, value, traceback)


sys.excepthook = exception_handler


def parse_args(config: GptCliConfig):
    parser = argparse.ArgumentParser(
        description="Run a chat session with ChatGPT. See https://github.com/grid-link-inc/gpt-cli for more information."
    )
    parser.add_argument(
        "assistant_name",
        type=str,
        default=config.default_assistant,
        nargs="?",
        choices=list(set([*DEFAULT_ASSISTANTS.keys(), *config.assistants.keys()])),
        help="The name of assistant to use. Name must match an assistant in the config file ~/.config/gpt-cli/gpt.yml. See the README for more information.",
    )
    parser.add_argument(
        "--no_markdown",
        action="store_false",
        dest="markdown",
        help="Disable markdown formatting in the chat session.",
        default=config.markdown,
    )
    parser.add_argument(
        "--log_file",
        type=str,
        default=config.log_file,
        help="The file to write logs to. Supports strftime format codes.",
    )
    parser.add_argument(
        "--log_level",
        type=str,
        default=config.log_level,
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="The log level to use",
    )
    parser.add_argument(
        "--no_stream",
        action="store_true",
        default=False,
        help="If specified, will not stream the response to standard output. This is useful if you want to use the response in a script. Ignored when the --prompt option is not specified.",
    )
    parser.add_argument(
        "--no_price",
        action="store_false",
        dest="show_price",
        help="Disable price logging.",
        default=config.show_price,
    )
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=f"gpt-cli v{gptcli.__version__}",
        help="Print the version number and exit.",
    )

    return parser.parse_args()


def main():
    config_file_path = choose_config_file(CONFIG_FILE_PATHS)
    if config_file_path:
        config = read_yaml_config(config_file_path)
    else:
        config = GptCliConfig()
    args = parse_args(config)

    if not config.api_key or not config.openai_api_key:
        print(
            "No API key found. Please set the OPENAI_API_KEY environment variable or `api_key: <key>` value in ~/.config/gpt-cli/gpt.yml"
        )
        sys.exit(1)

    assistant = init_assistant(cast(AssistantGlobalArgs, args), config.assistants)
    run_interactive(args, assistant)

class CLIChatSession(ChatSession):
    def __init__(self, assistant: AssistantThread, markdown: bool, show_price: bool):
        listeners = [
            CLIChatListener(markdown),
            LoggingChatListener(),
            PersistChatListener(assistant.get_thread_id(), assistant.get_assistant_id()),
        ]

        # TODO: Implement price for chatgpt Assistants
        # if show_price:
        #     listeners.append(PriceChatListener(assistant))

        listener = CompositeChatListener(listeners)
        super().__init__(assistant, listener)


def run_interactive(args, assistant):
    session = CLIChatSession(
        assistant=assistant, markdown=args.markdown, show_price=args.show_price
    )
    history_filename = os.path.expanduser("~/.config/gpt-cli/history")
    os.makedirs(os.path.dirname(history_filename), exist_ok=True)
    input_provider = CLIUserInputProvider(history_filename=history_filename)
    session.loop(input_provider)

if __name__ == "__main__":
    main()



"""
This module is responsible for reading the config file.
"""

import os
from typing import Dict, List, Optional
from attr import dataclass
import yaml

from gptcli.assistant import AssistantConfig


CONFIG_FILE_PATHS = [
    os.path.join(os.path.expanduser("~"), ".config", "gpt-cli", "gpt.yml"),
    os.path.join(os.path.expanduser("~"), ".gptrc"),
]


@dataclass
class GptCliConfig:
    default_assistant: str = ""
    markdown: bool = True
    show_price: bool = True
    api_key: Optional[str] = os.environ.get("OPENAI_API_KEY")
    openai_api_key: Optional[str] = os.environ.get("OPENAI_API_KEY")
    log_file: Optional[str] = None
    log_level: str = "INFO"
    assistants: Dict[str, AssistantConfig] = {}


def choose_config_file(paths: List[str]) -> str:
    for path in paths:
        if os.path.isfile(path):
            return path
    return ""


def read_yaml_config(file_path: str) -> GptCliConfig:
    with open(file_path, "r") as file:
        config = yaml.safe_load(file)
        return GptCliConfig(
            **config,
        )
